/* $Id: freedb.c,v 1.7 2005/05/11 19:53:15 matthew Exp $
 *
 * mcdp - command line CD playing
 * Copyright (C) 2003 Matthew Vernon <matthew@gnu.org>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Log: freedb.c,v $
 * Revision 1.7  2005/05/11 19:53:15  matthew
 * expose the get-discid function
 *
 * Revision 1.6  2005/05/11 19:40:14  matthew
 * add a function to return the discinfo structure from freedb
 *
 * Revision 1.5  2005/05/11 18:42:47  matthew
 * move the struct discinfo into the header file, so we can use it elsewhere
 *
 * Revision 1.4  2004/05/25 23:20:09  matthew
 * Tidy up output code
 *
 * Revision 1.3  2003/11/23 18:02:16  matthew
 * prettier output formatting
 *
 * Revision 1.2  2003/11/23 16:33:54  matthew
 * rudimentary CDDB support
 *
 * Revision 1.1  2003/11/23 14:47:59  matthew
 * split freedb functions out into separate .c file
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <errno.h>
#include <unistd.h>
#include <pwd.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <linux/cdrom.h>

#include "ferror.h"
#include "libmcdp.h"
#include "network.h"
#include "version.h"

struct discinfo *di;

static int cd_freedb_sum(int secs);
static unsigned char cd_freedb_query(const int drive, FILE *where);
static int cd_freedb_read_status_line(FILE *net, char *line);
static int cd_freedb_status(const char *line);
static struct discinfo *cd_freedb_lookup(const int drive, const int output);
static void cd_freedb_parse_error(const int code);
static void cd_freedb_parse_info (FILE *net, const char *discid, const unsigned char len, const int output);

/* Compute the discid for Freebd &c. Algorithm taken from 
 * http://www.freedb.org
 */
void cd_discid(const int drive)
{
  fprintf(stdout,"%08lx\n",cd_freedb_discid(drive));  
}

unsigned long cd_freedb_discid(const int drive)
{
  unsigned char l=0,a=0,b=0;
  unsigned long id;
  int n=0,start=0,t=0;
  struct cdrom_tocentry toc;
  toc.cdte_format=CDROM_MSF;

  l=cd_end_track(drive,CDROM_LAST_TRACK);
  b=cd_end_track(drive,CDROM_FIRST_TRACK);
  for(a=b;a<=l;a++){
    toc.cdte_track=(int)a;
    if(ioctl(drive,CDROMREADTOCENTRY,&toc )){
      fatal_error("ioctl CDROMREADTOCENTRY",1);
    }
    if(b==a)start=(toc.cdte_addr.msf.second+(60*toc.cdte_addr.msf.minute));
    n+=cd_freedb_sum(toc.cdte_addr.msf.second+(60*toc.cdte_addr.msf.minute));
  }
  /*Now deal with the lead-out track (0xAA) 
   *We use this for timing, but not for the "n" checksum
   */
  toc.cdte_track=0xAA;
  if(ioctl(drive,CDROMREADTOCENTRY,&toc )){
    fatal_error("ioctl CDROMREADTOCENTRY",1);
  }

  /*T is address of last track less address of first track*/
  t=((toc.cdte_addr.msf.second+(60*toc.cdte_addr.msf.minute))-start);
  /*Now compute the ID, remembering that l doesn't include the leadout*/
  id=((n % 0xff) << 24 | t << 8 | l);
  return(id);
}

/*algorithm from freedb.org again*/      
static int cd_freedb_sum(int secs)
{
  int ret=0;

  while (secs>0){
    ret=ret+(secs % 10);
    secs=secs/10;
  }

  return(ret);
}

unsigned char cd_output_query(const int drive)
{
  return(cd_freedb_query(drive, stdout));
}

static unsigned char cd_freedb_query(const int drive, FILE *where)
{
  unsigned long id,frames;
  unsigned char a=0,z=0,i;
  struct cdrom_tocentry toc;
  toc.cdte_format=CDROM_MSF;

  id=cd_freedb_discid(drive);
  a=cd_end_track(drive,CDROM_FIRST_TRACK);
  z=cd_end_track(drive,CDROM_LAST_TRACK);

  fprintf(where,"%08lx %u ",id,z);
  for(i=a;i<=z;i++){
    toc.cdte_track=(int)i;
    if(ioctl(drive,CDROMREADTOCENTRY,&toc )){
      fatal_error("ioctl CDROMREADTOCENTRY",1);
    }
    /*There are CD_FRAMES frames a second, and 60 seconds a minute*/
    frames=toc.cdte_addr.msf.frame;
    frames+=(toc.cdte_addr.msf.second * CD_FRAMES);
    frames+=(toc.cdte_addr.msf.minute * CD_FRAMES * 60);
    fprintf(where,"%lu ",frames);
  }
  /*length of CD==leadout offset in frames devide by 75*/
  toc.cdte_track=CDROM_LEADOUT;
  if(ioctl(drive,CDROMREADTOCENTRY,&toc )){
    fatal_error("ioctl CDROMREADTOCENTRY",1);
  }
  frames=toc.cdte_addr.msf.frame;
  frames+=(toc.cdte_addr.msf.second * CD_FRAMES);
  frames+=(toc.cdte_addr.msf.minute * CD_FRAMES * 60);
  fprintf(where,"%lu\n",(frames/75));

  return(z);
}

void cd_lookup_disc(const int drive)
{
  cd_freedb_lookup(drive,1);
}

struct discinfo *cd_get_discinfo(const int drive)
{
  return(cd_freedb_lookup(drive,0));
}

static struct discinfo *cd_freedb_lookup(const int drive, const int output)
{
  FILE *net;
  unsigned char len;
  char line[1024], *a, *b, *c, *id;
  int s;
  uid_t u;
  struct passwd *p;

  if(NULL==(di=malloc(sizeof(*di))))
    fatal_error("Unable to allocate memory for discinfo structure",1);

  net=cd_net_connect(FREEDB_HOSTNAME);

  s=cd_freedb_read_status_line(net,line);

  switch(s){
  case 200:
  case 201:
    break;
  default:
    cd_freedb_parse_error(s);
  }

  u=getuid();
  if(NULL==(p=(getpwuid(u))))
    fatal_error("getpwuid failed",1);
  if(-1==(gethostname(&(line[0]),1023)))
    fatal_error("gethostname failed",1);

  fprintf(net,"cddb hello %s %s mcdp %s\n",p->pw_name,line,VERSION);


  s=cd_freedb_read_status_line(net,line);
  switch(s){
  case 200:
    break;
  case 402:
    fprintf(stderr,"Warning: Server reports already shaken hands!\n");
    break;
  case 431:
    fatal_error("Handshake failed",0);
    break;
  default:
    cd_freedb_parse_error(s);
  }

  fprintf(net,"cddb query ");
  len=cd_freedb_query(drive, net);

  s=cd_freedb_read_status_line(net,line);
  switch(s){
  case 200:
    break;
  case 211:
    fatal_error("FIXME: support for inexact matches not implemented",0);
    break;
  case 202:
    fatal_error("No match for disc found\n",0);
    break;
  default:
    cd_freedb_parse_error(s);
  }

  /*parse the query reply - it is 
   *code categ discid dtitle
   *so we position a to be the beginning of categ, 
   *and c to be the space at the end of discid
   *(b is the beginning of discid)
   */
  if(NULL==(a=(index(line,' '))))
    fatal_error("Parsing query response",0);
  a++;
  if(NULL==a)
    fatal_error("1-word query response!",0);
  if(NULL==(b=(index(a,' '))))
    fatal_error("Parsing query response",0);
  b++;
  if(NULL==a)
    fatal_error("2-word query response!",0);
  if(NULL==(c=(index(b,' '))))
    fatal_error("Parsing query response",0);
  *c='\0';

  if(NULL==(id=(strdup(b))))
    fatal_error("Duplicating discid in memory",1);

  fprintf(net,"cddb read %s\n",a);

  s=cd_freedb_read_status_line(net,line);
  switch(s){
  case 210:
    break;
  default:
    cd_freedb_parse_error(s);
  }

  cd_freedb_parse_info(net,id,len,output);
  fprintf(net,"quit\n");
  return(di);

}

static int cd_freedb_read_status_line(FILE *net, char *line)
{

  if(NULL==fgets(line,1024,net)){
    if(ferror(net))
      fatal_error("Reading header from server",1);
    else fatal_error("No data from server",0);
  }
  
  return(cd_freedb_status(line));
}


static int cd_freedb_status(const char *line)
{
  long l;
  l=strtol(line,NULL,10);
  if(((LONG_MIN==l)||(LONG_MAX==l))&&(ERANGE==errno))
    fatal_error("status code out of range(!)",1);
  if(l<(0))
    fatal_error("status code was negative(!)",1);
  return(int)l;
}

static void cd_freedb_parse_error(const int code)
{
  switch(code){
  case 401:
    fatal_error("cddb read reports entry not found!",0);
    break;
  case 402:
    fatal_error("cddb server reports server error",0);
    break;
  case 403:
    fatal_error("cddb server reports database entry corrupt",0);
    break;
  case 408:
    fatal_error("cddb server reports CGI environment error",0);
    break;
  case 409:
    fatal_error("Server reports no handshake!",0);
    break;
  case 431:
    fatal_error("Handshake failed",0);
    break;
  case 432:
    fatal_error("Unable to connect to server: permission denied",0);
    break;
  case 433:
    fatal_error("Unable to connect to server: too many users",0);
    break;
  case 434:
    fatal_error("Unable to connect to server: system load too high",0);
    break;
  case 500:
    fatal_error("cddb server reports Command syntax error, command unknown, command unimplemented",0);
    break;
  case 530:
    fatal_error("cddb server error - server timeout",0);
    break;
  default:
    fatal_error("unrecognised response code",0);
  }
}

static void cd_freedb_parse_info (FILE *net, const char *discid, const unsigned char len, const int output)
{
  char line[1024];
  char *a;
  int i;

  if(NULL==(di->tracks=((char **)malloc((int)len*(sizeof(char *))))))
    fatal_error("allocating memory for track-name array",1);
  
  di->ntracks=(int)len;

  for(;;){
    if(NULL==fgets(line,1024,net)){
      if(ferror(net))
	fatal_error("Reading header from server",1);
      else break; /*EOF*/
    }
    /*If this is the last line, then quit*/
    if('.'==line[0]) break;
    else{
      if(!strncmp(line,"DTITLE=",7)){
	a=index(line,'=');
	a++;
	if(a){
	  if(NULL==(di->title=(strdup(a))))
	    fatal_error("allocating memory for title",0);
	}else{
	  if(NULL==(di->title=(strdup("[none]"))))
	    fatal_error("allocating memory for title",0);
	}
      }
      else if(!strncmp(line,"TTITLE",6)){
	i=cd_freedb_status(&(line[6]));
	if(i>(((int)len)-1))
	  fatal_error("Track data for too hight a track number",0);
	a=index(line,'=');
	a++;
	if(a){
	  if(NULL==(di->tracks[i]=(strdup(a))))
	    fatal_error("allocating memory for track title",0);
	}else{
	  if(NULL==(di->tracks[i]=(strdup("[none]"))))
	    fatal_error("allocating memory for track title",0);
	}
      }
    }
  }
  if(output){
    fprintf(stdout,"Title: %s",di->title);
    for(i=0;i<di->ntracks;i++)
      fprintf(stdout,"%2d. %s",i+1,di->tracks[i]);
  }
}

